/**
  *  \file client/dlg-cargo.cc
  *  \brief Cargo transfer window.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-cargo.h"
#include "client/cscreen.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "ui/window.h"
#include "ui/layout.h"
#include "ui/event.h"
#include "ui/msgbox.h"
#include "cpluslib/format.h"
#include "ui/util.h"
#include "cpluslib/holder.h"
#include "ui/skin.h"
#include "cpluslib/smallset.h"
#include "ui/prefix.h"
#include "ui/widgets/focusiterator.h"
#include "resmgr/resmgr.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "client/widgets/helpwidget.h"
#include "game/cargotr.h"
#include "ui/widgets/fixedspacer.h"
#include "util/misc.h"
#include "ui/widgets/quit.h"

namespace {
    class WOverloadMarker : public UIFixedWidget {
        int counter;
        bool state;
     public:
        WOverloadMarker();
        void drawContent(GfxCanvas& can);
        bool handleEvent(const UIEvent& e, bool second_pass);
    };

    class WCargoTransferWindow : public UIWindow {
        GCargoTransfer& transfer;
        WOverloadMarker overload_marker;
     public:
        WCargoTransferWindow(GCargoTransfer& transfer);
        ~WCargoTransferWindow();

        GCargoTransfer& getActor()
            { return transfer; }

        bool handleEvent(const UIEvent& e, bool second_pass);

        bool isSup2mc() const;

        void enableOverload();

        UIWidget* sup2mc;
    };

    class WCargoLine : public UIBaseComplexWidget {
        GCargoType       type;
        GCargoTransfer&  actor;
        UIButton         le, ri;
        void  showText(GfxContext& c, int side, int x, int y);
        void  doLayout();
     public:
        WCargoLine(int id, GCargoType type, GCargoTransfer& actor);
        void  drawContent(GfxCanvas& can);
        void  getLayoutInfo(LayoutInfo& info);
        bool  handleEvent(const UIEvent& e, bool second_pass);
        void  onResize(GfxRect oldsize, bool by_layout);
        void  onStateChange(int state, bool enable);
    };

    class WCargoHeader : public UIWidget {
        Ptr<GfxPixmap> header;
        void  showTitle(GfxCanvas& can, GCargoContainer& con, int x);
     public:
        WCargoHeader();
        void  drawContent(GfxCanvas& can);
        void  getLayoutInfo(LayoutInfo& info);
    };
};

/** \class WCargoHeader
    \brief Header line (two green boxes)

    A WCargoHeader displays the header with the names of the cargo
    transfer partners, plus the column headings

    \inv owner must be a WCargoTransferWindow */
WCargoHeader::WCargoHeader()
    : UIWidget(),
      header(requestPixmap("greentile"))
{
    setState(st_Disabled, true);
}

void
WCargoHeader::showTitle(GfxCanvas& can, GCargoContainer& con, int x)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(COLOR_WHITE);

    GfxRect toprect(x-10, extent.y, 190, font_heights[FONT_NORMAL]*3-3);

    drawTiledArea(can, toprect, header,
                  standard_colors[COLOR_GREENSCALE+4], 0 /* alteration */);
    drawFrameDown(can, toprect);

    outTextF(ctx, x, extent.y + 5, 180, con.getName());
    drawHLine(ctx, x, extent.y + font_heights[FONT_NORMAL] + 4, x + 180);

    ctx.setTextAlign(1, 0);
    const int y = extent.y + font_heights[FONT_NORMAL] + 6;
    outText(ctx, x+45, y, _("has"));
    outText(ctx, x+135, y, _("space left"));
}

void
WCargoHeader::drawContent(GfxCanvas& can)
{
    WCargoTransferWindow& w = *checked_cast<WCargoTransferWindow*>(getParent());
    showTitle(can, w.getActor()[0], extent.x + 10);
    showTitle(can, w.getActor()[1], extent.x + 430);
    can.update(getExtent());
}

void
WCargoHeader::getLayoutInfo(LayoutInfo& info)
{
    // FIXME: width is same as WCargoLine
    info.flags = li_FlexibleH;
    info.min_size = info.pref_size = GfxPoint(430+190, 3*font_heights[FONT_NORMAL]);
}

/******************************* WCargoLine ******************************/

static void
moveMax(GCargoTransfer& actor, GCargoType type, int from, bool s2m)
{
    int32_t visible = actor[from].getCargo(type) + actor[from].getDisplayOffset(type);
    if (visible <= 0) {
        visible = actor[from].getCargo(type);
    }
    actor.move(type, visible, from, true, s2m);
}

/** \class WCargoLine
    \brief One line in the cargo window

    Displays amounts for one cargo type, as well as two buttons for
    manipulating it. */

/** Constructor.
    \param id      widget id
    \param type    type of cargo
    \param actor   underlying transaction class */
WCargoLine::WCargoLine(int id, GCargoType type, GCargoTransfer& actor)
    : UIBaseComplexWidget(id), type(type), actor(actor),
      le("\xE2\x86\x90", UIKey(SDLK_LEFT)),
      ri("\xE2\x86\x92", UIKey(SDLK_RIGHT))
{
    addChildWidget(le, 0);
    addChildWidget(ri, 0);
    doLayout();
}

void
WCargoLine::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setColor(UIColor::tc_Static);
    ctx.useFont(*fonts[FONT_NORMAL]);

    string_t unit = getCargoUnit(type);
    if (unit.length())
        unit = " [" + unit + "]";

    outTextF(ctx, extent.x + 230, extent.y + 2, 170, getCargoName(type) + unit);

    ctx.setSolidBackground();
    showText(ctx, 0, extent.x + 10,  extent.y + 2);
    showText(ctx, 1, extent.x + 430, extent.y + 2);

    if (getFocusState() != NoFocus) {
        ctx.setColor(UIColor::tc_Contrast);
        drawRectangle(ctx, extent);
    } else {
        drawBackground(ctx, GfxRect(extent.x, extent.y, extent.w, 1));
        drawBackground(ctx, GfxRect(extent.x, extent.y+extent.h-1, extent.w, 1));
        drawBackground(ctx, GfxRect(extent.x, extent.y+1, 1, extent.h-2));
        drawBackground(ctx, GfxRect(extent.x+extent.w-1, extent.y+1, 1, extent.h-2));
    }

    can.update(extent);
}

void
WCargoLine::showText(GfxContext& ctx, int side, int x, int y)
{
    int32 now = actor[side].getCargo(type);
    int32 max = actor[side].getMaxCargo(type);

    ctx.setColor(UIColor::tc_Green);
    ctx.setTextAlign(2, 0);
    outTextF(ctx, x + 90, y, 90, numToString(now + actor[side].getDisplayOffset(type)));

    if (max > 20000)
        outTextF(ctx, x + 180, y, 90, _("(unl)"));
    else
        outTextF(ctx, x + 180, y, 90, numToString(max - now));
}

/* Display:

   Item              PCC1                     PCC2

   left              +19                      +10
                       +0  W90   amount        +0 W90
                       +90 W90   maximum       +90 W90
   button            +210  W19                +200 W20
   good name/unit    +240                     +230
   button            +410  W19                +400 W20
   right             +440                     +430          */
void
WCargoLine::doLayout()
{
    le.setExtent(GfxRect(extent.x + 200, extent.y + 2, 20, 16), true);
    ri.setExtent(GfxRect(extent.x + 400, extent.y + 2, 20, 16), true);
}

void
WCargoLine::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(430+190, 20);
}

bool
WCargoLine::handleEvent(const UIEvent& e, bool second_pass)
{
    // reject keypresses to inactive line
    if (e.type == ev_Key && !hasState(st_Focused))
        return false;

    // accept some other keypresses
    if (e.type == ev_Key && hasState(st_Focused)) {
        WCargoTransferWindow& w = *checked_cast<WCargoTransferWindow*>(getParent());
        bool s2m = w.isSup2mc();

        switch(e.key.code) {
         case SDLK_LEFT:
         case SDLK_RIGHT:
            {
                WCargoTransferWindow& w = *checked_cast<WCargoTransferWindow*>(getParent());
                bool s2m = w.isSup2mc();
                int dir = (e.key.code == SDLK_LEFT);
                int amount = e.key.arg ? e.key.arg : 10;
                actor.move(type, amount, dir, true, s2m);
                return true;
            }
         case ss_Shift + SDLK_LEFT:
            actor.move(type, 1, 1, true, s2m);
            return true;
         case ss_Shift + SDLK_RIGHT:
            actor.move(type, 1, 0, true, s2m);
            return true;
         case ss_Ctrl + SDLK_LEFT:
            actor.move(type, 100, 1, true, s2m);
            return true;
         case ss_Ctrl + SDLK_RIGHT:
            actor.move(type, 100, 0, true, s2m);
            return true;
         case ss_Alt + SDLK_LEFT:
            if (int n = e.key.arg)
                actor.move(type, n - actor[0].getCargo(type), 1, true, s2m);
            else
                moveMax(actor, type, 1, s2m);
            return true;
         case ss_Alt + SDLK_RIGHT:
            if (int n = e.key.arg)
                actor.move(type, n - actor[1].getCargo(type), 0, true, s2m);
            else
                moveMax(actor, type, 0, s2m);
            return true;
        }
    }

    // handle mouse clicks
    if (e.type == ev_Mouse && e.mouse.b && !hasState(st_Focused) && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        setSelectedWidget();
        setFocusedWidget();
        /* FALLTHROUGH */
    }
    return UIBaseComplexWidget::handleEvent(e, second_pass);
}

void
WCargoLine::onResize(GfxRect /*oldsize*/, bool by_layout)
{
    doLayout();
    if (!by_layout)
        drawWidget();
}

void
WCargoLine::onStateChange(int state, bool enable)
{
    if (state & st_Focused)
        drawWidget();
    UIBaseComplexWidget::onStateChange(state, enable);
}

/**************************** WOverloadMarker ****************************/

WOverloadMarker::WOverloadMarker()
    : counter(0), state(true)
{ }

void
WOverloadMarker::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setColor(state ? COLOR_WHITE : COLOR_BLACK);
    outTextF(ctx, extent.x, extent.y, extent.w, _("Overload Mode"));
    can.update(extent);
}

bool
WOverloadMarker::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Tick) {
        if (++counter > 40) {
            counter = 0;
            state = !state;
            drawWidget(false);
        }
    }
    return UIFixedWidget::handleEvent(e, second_pass);
}

/************************** WCargoTransferWindow *************************/

/** \class WCargoTransferWindow
    \brief Cargo Transfer Window (normal transfer) */

/** Constructor.
    \param left,right cargo transfer traits */
WCargoTransferWindow::WCargoTransferWindow(GCargoTransfer& transfer)
    : UIWindow(_("Cargo Transfer"), BLUE_WINDOW, COLOR_GRAY),
      transfer(transfer),
      sup2mc(0)
{
    setLayout(UIVBoxLayout::instance0);
}

WCargoTransferWindow::~WCargoTransferWindow()
{ }

bool
WCargoTransferWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case 's':
            if (sup2mc != 0) {
                sup2mc->setState(UIButton::bf_Highlight, !sup2mc->hasState(UIButton::bf_Highlight));
            }
            return true;

         case 'u':
            transfer.doUnload(isSup2mc());
            return true;

         case 'o' + ss_Alt:
            enableOverload();
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

bool
WCargoTransferWindow::isSup2mc() const
{
    return sup2mc && sup2mc->hasState(UIButton::bf_Highlight);
}

void
WCargoTransferWindow::enableOverload()
{
    /* already enabled? */
    if (transfer[0].hasFlag(GCargoContainer::can_Overload))
        return;

    /* ask user */
    if (!UIMessageBox(_("Overload Mode allows you to load more cargo on ships "
                       "than PCC usually permits. There are a few cases "
                       "where this is allowed and useful, otherwise the Host "
                       "will detect this and destroy your cargo.\n"
                       "Turn on Overload Mode?"),
                      _("Cargo Transfer")).doYesNoDialog())
        return;

    /* ok, enable it */
    for (int i = 0; i < 2; ++i) {
        GCargoContainer& c = transfer[i];
        c.setFlags(c.getFlags() | GCargoContainer::can_Overload);
    }
    getActor().sig_changed.raise();

    overload_marker.setExtent(GfxRect(extent.x + 5, extent.y + 2,
                                      extent.w/2, 24 /* FIXME hardcoded values */));
    add(overload_marker);
}

/** The order in which cargo types appear on the window. This is the
    same order as used in PCC 1.x. */
static const GCargoType elements[] = {
    el_Neutronium,  el_Tritanium,  el_Duranium,  el_Molybdenum,
    el_Supplies,    el_Colonists,  el_Money,
    el_Fighters,
    el_Torps,
    GCargoType(el_Torps+1),
    GCargoType(el_Torps+2),
    GCargoType(el_Torps+3),
    GCargoType(el_Torps+4),
    GCargoType(el_Torps+5),
    GCargoType(el_Torps+6),
    GCargoType(el_Torps+7),
    GCargoType(el_Torps+8),
    GCargoType(el_Torps+9)
};

/** Normal cargo transfer, main entry point. Does all of the cargo
    transfer actions between left and right. By convention, left is
    the object the user called the cargo transfer from. */
void
doCargoTransfer(GCargoTransfer& transfer)
{
    GCargoContainer& left  = transfer[0];
    GCargoContainer& right = transfer[1];

    /* figure out what can be transferred */
    GCargoTypeSet available_types;
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i)
        if (left.canHaveCargo(i) && right.canHaveCargo(i) && (left.getCargo(i) || right.getCargo(i)))
            available_types += i;
    if (available_types.contains(el_Supplies)
        && ((left.getFlags() | right.getFlags()) & GCargoContainer::can_Sup2mc))
        available_types += el_Money;

    if (available_types.empty()) {
        messageBox(format(_("There is nothing you could transfer to or from %s."),
                          right.getName()),
                   _("Cargo Transfer"));
        return;
    }

    /* create window */
    ObjectHolder holder;
    WCargoTransferWindow win(transfer);
    win.add(holder.add(new WCargoHeader()));
    win.add(holder.add(new UIPrefixArg()));
    int line_id = 1000;
    for (unsigned i = 0; i < countof(elements); ++i) {
        // WARNING: WCargoHeader and WCargoLine both assume that their immediate
        // parent is WCargoTransferWindow.
        if (available_types.contains(elements[i])) {
            UIBaseWidget& l = holder.add(new WCargoLine(line_id++, elements[i], win.getActor()));
            l.observe(win.getActor().sig_changed);
            win.add(l);
        }
    }
    win.add(holder.add(new UIFocusIterator(1000, line_id-1, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Wrap)));
    win.add(holder.add(new UIFixedSpacer(GfxPoint(5, 5))));

    UIGroup button_list(UIHBoxLayout::instance5);

    // FIXME: this layout is rather minimalistic and does not match PCC 1.x
    // PCC 1.x has
    //     [Sup2mc]      [OK] [Cancel] [H]    [<-Unload]
    // or  [Unload->]    [OK] [Cancel] [H]      [Sup2mc]
    // where both Sup2mc and Unload are optional, and the buttons are aligned
    // with the WCargoLine controls. This is hard to express using standard
    // layout managers, we probably need one specially-made for this situation.
    if (left.hasFlag(GCargoContainer::can_Sup2mc) || right.hasFlag(GCargoContainer::can_Sup2mc)) {
        win.sup2mc = &holder.add(new UIButton(_("Sup \xE2\x86\x92 mc"), 's'));
    }

    if (left.hasFlag(GCargoContainer::can_Unload) && right.hasFlag(GCargoContainer::can_GetUnload)) {
        /* [Sup2mc] [Unload->] [....] [OK] [Cancel] */
        if (win.sup2mc)
            button_list.add(*win.sup2mc);
        button_list.add(holder.add(new UIButton(_("Unload \xE2\x86\x92"), 'u')));
    } else if (right.hasFlag(GCargoContainer::can_Unload) && left.hasFlag(GCargoContainer::can_GetUnload)) {
        /* [Unload<-] [Sup2mc] [....] [OK] [Cancel] */
        button_list.add(holder.add(new UIButton(_("\xE2\x86\x90 Unload"), 'u')));
        if (win.sup2mc)
            button_list.add(*win.sup2mc);
    } else {
        /* [Sup2mc] [....] [OK] [Cancel] */
        if (win.sup2mc)
            button_list.add(*win.sup2mc);
    }
    button_list.add(holder.add(new UISpacer()));
    button_list.add(holder.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    button_list.add(holder.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    button_list.add(holder.add(new UIButton("H", 'h')));
    win.add(button_list);
    win.add(holder.add(new UIQuit(0)));
    win.add(holder.add(new WHelpWidget("pcc2:cargo")));

    /* run it */
    win.pack();
    ui_root->moveWidgetToEdge(win, 1, 0, 0);
    int rv = ui_root->runModal(win);

    if (rv)
        transfer.doCommit();
    else
        transfer.doCancel();
}
