/**
  *  \file client/labels.cc
  *  \brief Object Label Expressions
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include <memory>
#include "client/labels.h"
#include "cpluslib/sigslot.h"
#include "game/config.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/pref.h"
#include "game/ship.h"
#include "int/bytecode.h"
#include "int/exec.h"
#include "int/expr.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/interror.h"
#include "int/parseexpr.h"
#include "int/simple.h"
#include "client/history.h"
#include "int/process.h"

/** Stores all object labels. */
static std::vector<string_t> labels[2];

/** Compiled label expressions. */
static IntBCORef label_exprs[2];

/** Configuration file options for label expressions. */
static ConfigStringOption label_ship(getUserPreferences(), "Label.Ship");
static ConfigStringOption label_planet(getUserPreferences(), "Label.Planet");
static ConfigStringOption* const label_cfg[2] = { &label_ship, &label_planet };

static SignalConnection conn_turn;
static SignalConnection conn_edit;

/** Recompute labels.
    \param expr        [in] Expression to compute.
    \param labels      [out] Storage for labels. Must have sufficient size.
    \param what        [in] what to recompute.
    \param everything  [in] true to recompute all labels, false to recompute only those
                       whose object changed.
    \param trackErrors [in] true to track compilation errors; throws IntError when the
                       expression cannot be evaluated. false to ignore errors. */
static void
recomputeLabels(const IntBCORef& expr, std::vector<string_t>& labels, const LabelKind what, const bool everything, const bool trackErrors)
{
    string_t last_error;
    int num_evals = 0, num_errors = 0;

    /* Planets */
    if (what == PlanetLabel) {
        for (int pid = 1; pid <= NUM_PLANETS; ++pid) {
            /* Check whether we have to recompute */
            GPlanet& p = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
            if (!everything && !p.isDirty())
                continue;

            /* Recompute */
            labels[pid-1].clear();
            if (p.isVisible() && expr != 0) {
                ++num_evals;
                IntExecutionContext exec("Temporary: Labels");
                exec.pushNewContext(new IntPlanetContext(pid));
                exec.pushNewFrame(new IntExecutionFrame(expr));
                if (runTemporaryProcess(exec)) {
                    labels[pid-1] = IntValue::toStringOf(exec.getResult(), false);
                } else if (exec.getState() == IntExecutionContext::psError) {
                    last_error = exec.getError().what();
                    ++num_errors;
                } else {
                    // ignore silently
                }
            }
        }
    }

    /* Ships */
    if (what == ShipLabel) {
        for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
            /* Check whether we have to recompute */
            GShip& s = getDisplayedTurn().getCurrentUniverse().getShip(sid);
            if (!everything && !s.isDirty())
                continue;

            /* Recompute */
            labels[sid-1].clear();
            if (s.isVisible() && expr != 0) {
                ++num_evals;
                IntExecutionContext exec("Temporary: Labels");
                exec.pushNewContext(new IntShipContext(sid));
                exec.pushNewFrame(new IntExecutionFrame(expr));
                if (runTemporaryProcess(exec)) {
                    labels[sid-1] = IntValue::toStringOf(exec.getResult(), false);
                } else if (exec.getState() == IntExecutionContext::psError) {
                    last_error = exec.getError().what();
                    ++num_errors;
                } else {
                    // ignore silently
                }
            }
        }
    }

    /* If all our evaluations failed, throw an error */
    if (trackErrors && num_errors > 0 && num_errors == num_evals) {
        throw IntError(last_error);
    }
}

/** Universe change: called when an object (may) change in the current universe.
    Recomputes all changed labels. */
static void
onUniverseChanged()
{
    recomputeLabels(label_exprs[PlanetLabel], labels[PlanetLabel], PlanetLabel, false, false);
    recomputeLabels(label_exprs[ShipLabel],   labels[ShipLabel],   ShipLabel,   false, false);
}

/** Turn change: called when a new turn is displayed. Recompute everything and watch new turn. */
static void
onTurnChanged()
{
    recomputeLabels(label_exprs[PlanetLabel], labels[PlanetLabel], PlanetLabel, true, false);
    recomputeLabels(label_exprs[ShipLabel],   labels[ShipLabel],   ShipLabel,   true, false);
    conn_edit = getDisplayedTurn().getCurrentUniverse().sig_pre_update.add(onUniverseChanged);
}

/** Initialize labels. Must be called after loading configuration,
    before displaying the map. */
static void
initLabels()
{
    labels[ShipLabel].clear();
    labels[ShipLabel].resize(NUM_SHIPS);
    labels[PlanetLabel].clear();
    labels[PlanetLabel].resize(NUM_PLANETS);

    label_exprs[ShipLabel]   = compileExpression((*label_cfg[ShipLabel])(), IntCompilationContext());
    label_exprs[PlanetLabel] = compileExpression((*label_cfg[PlanetLabel])(), IntCompilationContext());

    conn_turn = sig_turn_changed.add(onTurnChanged);
}

/** Get label.
    \param kind Kind of object to query
    \param id Object Id */
string_t
getLabel(LabelKind kind, int id)
{
    if (id > 0 && id <= int(labels[kind].size()))
        return labels[kind][id-1];
    else
        return string_t();
}

/** Get label expression.
    \param kind Kind of object to query */
string_t
getLabelExpr(LabelKind kind)
{
    return (*label_cfg[kind])();
}

/** Set label expression. This function is atomic in that it either sets
    the label expression successfully or does not change anything at all.
    \param kind Kind of object to query
    \param expr Expression
    \throw IntError if there is a problem with the expression */
void
setLabelExpr(LabelKind kind, string_t expr)
{
    IntTokenizer tok(expr);
    IntBCORef bco;
    if (tok.getCurrentToken() != tok.tEnd) {
        /* expression given, parse it */
        std::auto_ptr<IntExprNode> expr(IntExpressionParser(tok).parse());
        if (tok.getCurrentToken() != tok.tEnd)
            throw IntError::garbageAtEnd(true);

        bco = new IntBytecodeObject();
        expr->compileValue(*bco, IntCompilationContext());
        bco->relocate();
    }

    /* Try to compute the labels */
    std::vector<string_t> tmp(kind == PlanetLabel ? NUM_PLANETS : NUM_SHIPS);
    recomputeLabels(bco, tmp, kind, true, true);

    /* When we are here, label computation succeeded, so keep
       expression, labels, and BCO */
    labels[kind].swap(tmp);
    label_cfg[kind]->set(expr);
    label_exprs[kind] = bco;
}

// Constraints: after mod_pref, before setDisplayedTurn()
extern const ModuleInitializer mod_labels[] = {
    { OnEnterGame, initLabels },
    // FIXME { OnLeaveGame, doneLabels },
    { 0, 0 },
};
