/**
  *  \file client/widgets/funclist.cc
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/funclist.h"
#include "game/struct.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"

WHullFunctionList::WHullFunctionList()
    : UIListbox(0, 10, 30*font_ems[FONT_NORMAL], 0)
{ }

WHullFunctionList::~WHullFunctionList()
{ }

void
WHullFunctionList::setQuery(const GUniverse& univ, const GShipQuery& query)
{
    if (query != this->query) {
        // Rebuild list
        this->query = query;
        list.clear();
        query.enumShipSpecials(univ, list, true);
        list.simplify();
        list.sortForNewShip(query.getPlayerDisplaySet());

        // Update display
        setSize(list.size());
        setTop(0);
        setCurrent(0);
        sig_moved.raise(getId());
        drawWidget();
    }
}

void
WHullFunctionList::drawPart(GfxCanvas& can, int from, int to)
{
    const GfxRect& e = getExtent();

    GfxContext ctx;
    ctx.useCanvas(can);

    int w1 = fonts[FONT_NORMAL]->getTextWidth(" [U] ") + 3;
    int w2 = fonts[FONT_NORMAL]->getTextWidth("+ ") + 3;
    int w3 = e.w - w1 - w2;

    for (int i = from; i < to; ++i) {
        /* Prepare */
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);

        /* Each line consists of three segments:
           - Assignment indicator icon
               [U]   Universal ability (all ships, all races)
               [R]   Racial ability (all ships, some races)
               [G]   Global class function (this class, all races)
               [C]   Class function (this class, some races)
               [S]   Ship function (this ship)
           - Availability marker
               >     You have it (FIXME: filled triangle in PCC 1.x)
               +     Need higher level
               -     Level too high
               blank Other races only
           - Function name */
        int item = getTop() + i;

        string_t s1, s2, s3;
        ctx.useFont(*fonts[FONT_NORMAL]);
        if (item >= 0 && item < list.size()) {
            const GHullFunction& hf = list[item];
            /* Assignment indicator icon */
            if (hf.getKind() == GHullFunction::AssignedToRace) {
                if (hf.getPlayers().contains(GPlayerSet::allUpTo(NUM_PLAYERS) - 0))
                    s1 = " [U] ";
                else
                    s1 = " [R] ";
            } else if (hf.getKind() == GHullFunction::AssignedToHull) {
                if (hf.getPlayers().contains(GPlayerSet::allUpTo(NUM_PLAYERS) - 0))
                    s1 = " [G] ";
                else
                    s1 = " [C] ";
            } else {
                s1 = " [S] ";
            }

            /* Availability marker */
            if ((hf.getPlayers() & query.getPlayerDisplaySet()).empty()) {
                /* Not available */
                ctx.setColor(UIColor::tc_Faded);
            } else if ((hf.getLevels() & query.getLevelDisplaySet()).empty()) {
                /* Wrong experience */
                // FIXME: check this condition. It's the same as in PCC 1.x.
                if (hf.getLevels().toInteger() >= 2*query.getLevelDisplaySet().toInteger())
                    s2 = "+";
                else
                    s2 = "-";
                ctx.setColor(UIColor::tc_Blue);
            } else {
                /* OK */
                s2 = "\xE2\x96\xB6";
                ctx.useFont(*fonts[FONT_NORMAL_BOLD]);
            }

            /* Name */
            if (const GBasicHullFunction* ptr = GHullFunctionData::instance.getBasicFunctionById(hf.getBasicFunctionId()))
                s3 = ptr->getDescription();
            else
                s3 = format(_("Hull Function #%d"), hf.getBasicFunctionId());
        }

        /* Draw it */
        outTextF(ctx, area.x+w1+w2, area.y, w3, s3);
        ctx.useFont(*fonts[FONT_NORMAL]);
        outTextF(ctx, area.x,       area.y, w1, s1);
        outTextF(ctx, area.x+w1,    area.y, w2, s2);
    }
}

const GHullFunction*
WHullFunctionList::getSelectedFunction() const
{
    if (getCurrent() >= 0 && getCurrent() < list.size())
        return &list[getCurrent()];
    else
        return 0;
}
