/**
  *  \file game/sim-data.cc
  *  \brief Battle Simulator - Data
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/sim.h"
#include "game/config.h"
#include "game/hullfunc.h"
#include "util/recode.h"
#include "io/io.h"

        
static bool
checkHullFunction(const GSimShip& sh, int basic_function)
{
    /* FIXME: PCC 1.1.17 decides upon host version here, and checks ImperialAssault
       on older hosts instead of PlanetImmunity. Ideally, the hull function module
       would isolate us from these differences. */
    if (sh.getHull() <= 0 || sh.getHull() > getNumHulls()) {
        // Hull does not exist (e.g., custom ship)
        return false;
    } else {
        // Check hull
        return ::getHull(sh.getHull()).canDoSpecial(basic_function,
                                                    sh.getOwner(),
                                                    true,
                                                    GExpLevelSet(sh.getExperienceLevel()));
    }
}

const uint32_t GSimObject::simf_to_value_bit[] = {
    GSimObject::fl_PlanetImmunity,
    GSimObject::fl_FullWeaponry,
    GSimObject::fl_Commander,
    GSimObject::fl_TripleBeamKill,
    GSimObject::fl_DoubleBeamCharge,
    GSimObject::fl_DoubleTorpCharge,
    GSimObject::fl_Elusive,
    GSimObject::fl_Squadron,
};
const uint32_t GSimObject::simf_to_set_bit[] = {
    GSimObject::fl_PlanetImmunitySet,
    GSimObject::fl_FullWeaponrySet,
    GSimObject::fl_CommanderSet,
    GSimObject::fl_TripleBeamKillSet,
    GSimObject::fl_DoubleBeamChargeSet,
    GSimObject::fl_DoubleTorpChargeSet,
    GSimObject::fl_ElusiveSet,
    GSimObject::fl_SquadronSet,
};


/*! \class GSimObject
    \brief Base class for simulator objects

    Contains the definition and state of an object. In addition to the
    normal specification data, this includes some counters for
    statistics tracking: the simulator updates these counters the same
    way as it updates the shield level. */

/** Constructor. Create blank object. */
GSimObject::GSimObject()
    : id(1), damage(0), shield(100), owner(12), experience_level(0), flags(0),
      flak_r_override(0), flak_c_override(0)
{
    storeBasicString(name, "?");
    storeBasicString(fcode, "?" "?" "?");
}

/** Destructor. */
GSimObject::~GSimObject()
{ }

/** Get name of this object. */
string_t
GSimObject::getName() const
{
    return convertGameToUtf8(getBasicString(name));
}

/** Change object name. */
void
GSimObject::setName(string_t name)
{
    storeBasicString(this->name, convertUtf8ToGame(name));
    setChanged();
}

/** Get friendly code. */
string_t
GSimObject::getFCode() const
{
    return convertGameToUtf8(getBasicString(fcode));
}

/** Change friendly code. */
void
GSimObject::setFCode(string_t fcode)
{
    storeBasicString(this->fcode, convertUtf8ToGame(fcode));
    setChanged();
}

/** Get maximum permitted damage level. Considers owner race. Intended
    for display / editing only, so this may allow more than the host
    allows. */
int
GSimObject::getMaxDamage() const
{
    if (raceId(getOwner()) == 2)
        return 150;
    else
        return 99;
}

/** Get maximum permitted shield level. Considers owner race and
    current damage. Intended for display / editing only, so this may
    allow more than the host allows. */
int
GSimObject::getMaxShield() const
{
    int limit = getMaxDamage() - getDamage() + 1;
    if (limit > 100)
        limit = 100;
    return limit;
}

/** Assign random friendly code if requested. Considers fl_RandomFC
    and the fl_RandomDigits flags to assign a new, (partially) numeric
    friendly code. */
void
GSimObject::assignRandomFCode()
{
    if (flags & fl_RandomFC) {
        int which = flags & fl_RandomDigits;
        if (which == 0)
            which = fl_RandomDigits;
        for (int i = 0; i < 3; ++i)
            if (which & (fl_RandomFC1 << i))
                fcode[i] = '0' + (rand() % 10);
        setChanged();
    }
}

/** Assign random friendly code flags. Derives fl_RandomFC and fl_RandomDigits
    from the actual friendly code selected.
    \return true iff fl_RandomFC has been enabled */
bool
GSimObject::assignRandomFCodeFlags()
{
    int32_t new_flags = flags & ~(fl_RandomFC | fl_RandomDigits);
    for (int i = 0; i < 3; ++i)
        if (fcode[i] == '#')
            new_flags |= (fl_RandomFC1 << i);
    if ((new_flags & fl_RandomDigits) != 0)
        new_flags |= fl_RandomFC;

    setFlags(new_flags);

    return (new_flags & fl_RandomFC) != 0;
}

/** Check whether this unit has a specific function. Consults specification and user settings. */
bool
GSimObject::hasFunction(GSimFunction simf) const
{
    const int32_t value = simf_to_value_bit[simf];
    const int32_t set = simf_to_set_bit[simf];
    const int32_t flags = getFlags();
    if ((flags & set) != 0) {
        // User has explicitly configured this function
        return (flags & value) != 0;
    } else {
        // Use default from specification (derived-class function)
        return hasImpliedFunction(simf);
    }
}

/** Get this object as a ship.
    \return \c this if this is a ship, 0 otherwise. */
GSimShip*
GSimObject::asShip()
{
    return 0;
}

/** Get this object as a planet.
    \return \c this if this is a planet, 0 otherwise. */
GSimPlanet*
GSimObject::asPlanet()
{
    return 0;
}

/******************************** GSimShip *******************************/

/*! \class GSimShip
    \brief Ship Data for Simulator */

/** Construct blank simulator ship. */
GSimShip::GSimShip()
    : crew(10), hull(0), mass(100), beam_type(0), beam_count(0),
      torp_type(0), torp_lcount(0), bay_count(0), ammo(0),
      engine_type(NUM_ENGINES), aggressiveness(agg_Passive),
      intercept_id(0)
{ }

/** Construct from file data.
    \param file_data [in] Ship record read from a .ccb file
    \param version   [in] .ccb version (last digit of file magic) */
GSimShip::GSimShip(const TSimShipData& file_data, int version)
    : crew(file_data.vcro.crew),
      hull(file_data.hull_type),
      mass(file_data.mass),
      beam_type(file_data.vcro.beam_type),
      beam_count(file_data.vcro.beam_count),
      torp_type(file_data.vcro.torp_type),
      torp_lcount(file_data.vcro._torp_launcher_count),
      bay_count(file_data.vcro.bay_count),
      ammo(file_data.vcro._ammo),
      engine_type(file_data.engine_type),
      aggressiveness(file_data.aggressiveness),
      intercept_id(file_data.xa_id)
{
    setExperienceLevel(file_data.vcro.experience_level);
    setOwner(file_data.vcro.owner);
    setId(file_data.vcro.id);
    setDamage(file_data.vcro.damage);
    setShield(file_data.shield);
    setName(convertGameToUtf8(getBasicString(file_data.vcro.name)));
    setFCode(convertGameToUtf8(getBasicString(file_data.fcode)));
    if (version < 3) {
        setFlags(0);
    } else if (version < 5) {
        setFlags(uint16_t(file_data.flags));
    } else {
        setFlags(uint16_t(file_data.flags) + 65536*file_data.flags2);
    }
    setFlakRatingOverride(file_data.flak_r);
    setFlakCompensationOverride(file_data.flak_c);
    if (version <= 0) {
        setAggressiveness(agg_Kill);
    }
    if (version <= 3) {
        setFlags(getFlags() & ~fl_RatingOverride);
        setInterceptId(0);
    }
    if (hull > 0 && hull <= getNumHulls()) {
        setMass(::getHull(hull).getMass());
    }
}

GSimShip::~GSimShip()
{ }

/** Store data into external representation.
    \param file_data [out] Data
    \param version   [in] Version to store for. Optimize/adapt output for
                     this file format version. */
void
GSimShip::storeData(TSimShipData& file_data, int version)
{
    (void) version;

    storeBasicString(file_data.vcro.name, convertUtf8ToGame(getName()));
    file_data.vcro.damage               = getDamage();
    file_data.vcro.crew                 = getCrew();
    file_data.vcro.id                   = getId();
    file_data.vcro.owner                = getOwner();
    file_data.vcro.race_or_zero         = 0;   // we don't use this field
    file_data.vcro.picture              = 0;   // we don't use this field
    file_data.vcro.hull_or_zero         = 0;   // we don't use this field
    file_data.vcro.beam_type            = getBeamType();
    file_data.vcro.beam_count           = getBeamCount();
    file_data.vcro.experience_level     = getExperienceLevel();
    file_data.vcro.bay_count            = getBayCount();
    file_data.vcro.torp_type            = getTorpType();
    file_data.vcro._ammo                = getAmmo();
    file_data.vcro._torp_launcher_count = getTorpLauncherCount();
    file_data.engine_type               = getEngineType();
    file_data.hull_type                 = getHull();
    file_data.shield                    = getShield();
    storeBasicString(file_data.fcode, convertUtf8ToGame(getFCode()));
    file_data.aggressiveness            = getAggressiveness();
    file_data.mass                      = getMass();
    file_data.flags                     = getFlags();
    file_data.flak_r                    = getFlakRatingOverride();
    file_data.flak_c                    = getFlakCompensationOverride();
    file_data.xa_id                     = getInterceptId();
    file_data.flags2                    = getFlags() >> 16;
}

/** Change crew. */
void
GSimShip::setCrew(int crew)
{
    this->crew = crew;
    setChanged();
}

/** Set hull type. Can be 0 for custom ships. If this changes the hull
    type, it will also update the arms and crew to reflect that new
    type. */
void
GSimShip::setHull(int hull)
{
    if (hull != this->hull) {
        this->hull = hull;
        if (hull) {
            /* it's a real hull. Fill in defaults. */
            const GHull& def = ::getHull(hull);

            // beams
            this->beam_count = def.getMaxBeams();
            if (this->beam_count) {
                if (!this->beam_type)
                    this->beam_type = NUM_BEAMS;
            }

            // torps
            if (def.getNumBays()) {
                this->torp_lcount = this->torp_type = 0;
                this->bay_count = def.getNumBays();
                this->ammo = def.getCargo();
            } else if (def.getMaxLaunchers()) {
                this->torp_lcount = def.getMaxLaunchers();
                if (!this->torp_type)
                    this->torp_type = NUM_TORPS;
                this->ammo = def.getCargo();
                this->bay_count = 0;
            } else {
                this->torp_lcount = this->torp_type = 0;
                this->bay_count = 0;
                this->ammo = 0;
            }

            // rest
            this->crew = def.getCrew();
            this->mass = def.getMass();
        }
        setChanged();
    }
}

/** Change mass. Only used for custom ships. */
void
GSimShip::setMass(int mass)
{
    this->mass = mass;
    setChanged();
}

/** Change beam type. */
void
GSimShip::setBeamType(int beam_type)
{
    this->beam_type = beam_type;
    setChanged();
}

/** Change beam count. */
void
GSimShip::setBeamCount(int beam_count)
{
    this->beam_count = beam_count;
    setChanged();
}

/** Change torpedo type. */
void
GSimShip::setTorpType(int torp_type)
{
    this->torp_type = torp_type;
    setChanged();
}

/** Change torpedo launcher count. */
void
GSimShip::setTorpLauncherCount(int torp_lcount)
{
    this->torp_lcount = torp_lcount;
    setChanged();
}

/** Change number of fighter bays. */
void
GSimShip::setBayCount(int bay_count)
{
    this->bay_count = bay_count;
    setChanged();
}

/** Change ammunition (torps, fighters). */
void
GSimShip::setAmmo(int ammo)
{
    this->ammo = ammo;
    setChanged();
}

/** Change engine type. */
void
GSimShip::setEngineType(int engine_type)
{
    this->engine_type = engine_type;
    setChanged();
}

/** Get aggressiveness. */
int
GSimShip::getAggressiveness() const
{
    return aggressiveness;
}

/** Change aggressiveness (agg_XXX). */
void
GSimShip::setAggressiveness(int aggressiveness)
{
    this->aggressiveness = aggressiveness;
    setChanged();
}

/** Set Id for intercept-attack. */
void
GSimShip::setInterceptId(int id)
{
    this->intercept_id = id;
    setChanged();
}

GSimShip*
GSimShip::asShip()
{
    return this;
}

/** Check whether this ship matches the current ship list. */
bool
GSimShip::isMatchingShipList() const
{
    /* custom ships have full freedom */
    if (isCustomShip())
        return true;

    /* valid hull? */
    /* FIXME: we cannot handle these during simulation so we should
       avoid even loading them. */
    if (getHull() > getNumHulls())
        return false;
    const GHull& h = ::getHull(getHull());

    /* beams */
    if (getBeamCount() > h.getMaxBeams())
        return false;

    /* torps / fighters */
    if (getTorpLauncherCount() > h.getMaxLaunchers())
        return false;
    if (getBayCount() != h.getNumBays())
        return false;
    if (getAmmo() > h.getCargo())
        return false;

    return true;
}

/** Check whether this ship has a specific hull function from the configuration. */
bool
GSimShip::hasImpliedFunction(GSimFunction simf) const
{
    switch (simf) {
     case sf_PlanetImmunity:
        return checkHullFunction(*this, hf_PlanetImmunity)
            || (raceId(getOwner()) == 4 && !config.PlanetsAttackKlingons())
            || (raceId(getOwner()) == 10 && !config.PlanetsAttackRebels());
     case sf_FullWeaponry:
        return checkHullFunction(*this, hf_FullWeaponry);
     case sf_Commander:
        return checkHullFunction(*this, hf_Commander);
     case sf_TripleBeamKill:
        return raceId(getOwner()) == 5;
     case sf_DoubleBeamCharge:
     case sf_DoubleTorpCharge:
     case sf_Elusive:
     case sf_Squadron:
        return false;
    }
    return false;
}

/******************************* GSimPlanet ******************************/

/*! \class GSimPlanet
    \brief Planet Data for Simulator */

/** Constructor. Construct blank planet. */
GSimPlanet::GSimPlanet()
    : defense(10), base_defense(10), beam_tech(0),
      torp_tech(1), base_fighters(0)
{
    for (int i = 0; i < NUM_TORPS; ++i)
        base_torps[i] = 0;
}

/** Construct from file data.
    \param file_data [in] Record read from .ccb file
    \param version   [in] File version */
GSimPlanet::GSimPlanet(const TSimPlanetData& file_data, int version)
    : defense(file_data.defense),
      base_defense(file_data.base_defense),
      beam_tech(file_data.beam_tech),
      torp_tech(file_data.torp_tech),
      base_fighters(file_data.base_fighters)
{
    for (int i = 0; i < NUM_TORPS; ++i)
        base_torps[i] = 0;
    if (beam_tech < 0 || beam_tech > NUM_BEAMS)
        beam_tech = 0;
    if (torp_tech <= 0 || torp_tech > NUM_TORPS)
        torp_tech = 1;
    if (base_defense < 0 || base_defense > config.MaximumDefenseOnBase(file_data.owner))
        base_defense = 0;
    if (defense < 0 || defense > 1000 /* FIXME */)
        defense = 0;

    setExperienceLevel(file_data.experience_level);
    setOwner(file_data.owner);
    setId(file_data.id);
    setDamage(0); // FIXME?
    setShield(100); // FIXME?
    setName("Planet"); /* FIXME: use spec */
    setFCode(convertGameToUtf8(getBasicString(file_data.fcode)));
    if (version < 3) {
        setFlags(0);
    } else if (version < 5) {
        setFlags(uint16_t(file_data.flags));
    } else {
        setFlags(uint16_t(file_data.flags) + 65536*file_data.flags2);
    }
    setFlakRatingOverride(file_data.flak_r);
    setFlakCompensationOverride(file_data.flak_c);

    if (version > 1) {
        for (int i = 0; i < NUM_TORPS; ++i)
            base_torps[i] = file_data.base_torps[i];
    } else {
        base_torps[torp_tech-1] = file_data.base_torps_old;
    }
    if (version <= 3)
        setFlags(getFlags() & ~fl_RatingOverride);
}

/** Destructor. */
GSimPlanet::~GSimPlanet()
{ }

/** Store data into external representation.
    \param file_data [out] Data
    \param version   [in] Version to store for. Optimize/adapt output for
                     this file format version. */
void
GSimPlanet::storeData(TSimPlanetData& file_data, int version)
{
    if (version > 1) {
        for (int i = 0; i < NUM_TORPS; ++i)
            file_data.base_torps[i] = base_torps[i];
        file_data.base_torps_old = 0;
    } else {
        for (int i = 0; i < NUM_TORPS; ++i)
            file_data.base_torps[i] = 0;
        file_data.base_torps_old = (torp_tech > 0 && torp_tech <= NUM_TORPS ? base_torps[torp_tech-1] : 0);
    }

    file_data._pad0            = 0;
    file_data.id               = getId();
    file_data.owner            = getOwner();
    file_data._pad1            = 0;
    file_data.beam_tech        = getBaseBeamTech();
    file_data._pad2            = 0;
    file_data.experience_level = getExperienceLevel();
    file_data.base_fighters    = getBaseFighters();
    file_data._pad3            = 0;
    file_data.torp_tech        = getBaseTorpTech();
    file_data.base_defense     = getBaseDefense();
    file_data.defense          = getDefense();
    file_data.shield           = getShield();
    storeBasicString(file_data.fcode, convertUtf8ToGame(getFCode()));
    file_data.aggressiveness   = -1 /* getAggressiveness() */;
    file_data._pad5            = 0;
    file_data.flags            = getFlags();
    file_data.flak_r           = getFlakRatingOverride();
    file_data.flak_c           = getFlakCompensationOverride();
    file_data._pad6            = 0;
    file_data.flags2           = getFlags() >> 16;
}

/** Change number of planetary defense posts. */
void
GSimPlanet::setDefense(int defense)
{
    this->defense = defense;
    setChanged();
}

/** Change number of base defense posts. */
void
GSimPlanet::setBaseDefense(int base_defense)
{
    this->base_defense = base_defense;
    setChanged();
}

/** Change starbase beam tech level. Zero means no base. */
void
GSimPlanet::setBaseBeamTech(int beam_tech)
{
    this->beam_tech = beam_tech;
    setChanged();
}

/** Change starbase torpedo tech level. */
void
GSimPlanet::setBaseTorpTech(int torp_tech)
{
    this->torp_tech = torp_tech;
    setChanged();
}

/** Change number of starbase fighters. */
void
GSimPlanet::setBaseFighters(int base_fighters)
{
    this->base_fighters = base_fighters;
    setChanged();
}

/** Get number of starbase torps of a given type. */
int
GSimPlanet::getBaseTorps(int tech) const
{
    ASSERT(tech > 0 && tech <= NUM_TORPS);
    return base_torps[tech-1];
}

/** Change number of starbase torps of a given type. */
void
GSimPlanet::setBaseTorps(int tech, int amount)
{
    ASSERT(tech > 0 && tech <= NUM_TORPS);
    this->base_torps[tech-1] = amount;
    setChanged();
}

/** Get number of base torpedoes for a given tech. Computes the
    effective torpedo count from all storage torpedoes. */
int
GSimPlanet::getBaseTorpsAsType(int tech) const
{
    uint32 total_cost = 0;
    for (int i = 1; i <= NUM_TORPS; ++i)
        total_cost += getBaseTorps(i) * getTorp(i).getTorpCost();

    uint32 cost = getTorp(tech).getTorpCost();
    if (cost != 0)
        total_cost /= cost;
    // FIXME: some range checking here?
    return total_cost;
}

bool
GSimPlanet::hasImpliedFunction(GSimFunction simf) const
{
    switch (simf) {
     case sf_PlanetImmunity:
     case sf_FullWeaponry:
     case sf_Commander:
        return false;
     case sf_TripleBeamKill:
        return raceId(getOwner()) == 5;
     case sf_DoubleBeamCharge:
     case sf_DoubleTorpCharge:
     case sf_Elusive:
     case sf_Squadron:
        return false;
    }
    return false;
}

GSimPlanet*
GSimPlanet::asPlanet()
{
    return this;
}
