/**
  *  \file game/sim.cc
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <climits>
#include "game/sim.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/team.h"

/****************************** GSimOptions ******************************/

/*! \class GSimOptions
    \brief Simulator Settings */

/** Constructor. Creates a default object. */
GSimOptions::GSimOptions()
    : alliance_settings(),
      enemy_settings(),
      engine_shield_bonus(0),
      scotty_bonus(true),
      random_left_right(false),
      honor_alliances(true),
      only_one_simulation(false),
      seed_control(false),
      randomize_fcodes_on_every_fight(false),
      balancing_mode(BalanceNone),
      vcr_mode(VcrPHost4),
      seed(0)
{ }

bool
GSimOptions::isExperienceEnabled() const
{
    switch (vcr_mode) {
     case VcrPHost4:
     case VcrFLAK:
        return config.NumExperienceLevels() > 0;

     case VcrPHost2:
     case VcrPHost3:
     case VcrNuHost:
     case VcrHost:
        return false;
    }
    return false;
}

/** Load defaults for a particular host type. */
void
GSimOptions::loadDefaults(VcrMode mode)
{
    alliance_settings.clear();
    enemy_settings.clear();
    setMode(mode);
    honor_alliances     = true;
    only_one_simulation = false;
    seed_control        = false;
    randomize_fcodes_on_every_fight = false;
    seed                = 0;

    // Initialize alliances from teams:
    for (int a = 1; a <= NUM_PLAYERS; ++a)
        for (int b = 1; b <= NUM_PLAYERS; ++b)
            if (a != b && getPlayerTeam(a) != 0 && getPlayerTeam(a) == getPlayerTeam(b))
                alliance_settings.set(a, b, true);
}

void
GSimOptions::setMode(VcrMode mode)
{
    engine_shield_bonus = (config.AllowEngineShieldBonus() ? config.EngineShieldBonusRate(getPlayerId()) : 0);
    scotty_bonus        = config.AllowFedCombatBonus();
    random_left_right   = mode != VcrHost && mode != VcrNuHost;
    balancing_mode      = (mode != VcrHost && mode != VcrNuHost ? BalanceNone : Balance360k);
    vcr_mode            = mode;
}

/******************************* GSimState *******************************/

/*! \class GSimState
    \brief Simulator State

    This object carries a simulation set-up. Users can edit these
    objects; it provides methods for change tracking. In addition, the
    actual simulator operates on GSimState objects.

    GSimState can contain a number of ships and up to one planet. */

/** Create blank object. */
GSimState::GSimState()
    : has_planet(0)
{ }

/** Copy constructor. */
GSimState::GSimState(const GSimState& other)
    : ships(other.ships), planet(other.planet), has_planet(other.has_planet)
{ }

/** Destructor. */
GSimState::~GSimState()
{ }

/** Assignment. */
GSimState&
GSimState::operator=(const GSimState& other)
{
    ships = other.ships;
    planet = other.planet;
    has_planet = other.has_planet;
    structure_change = true;
    return *this;
}

/** Make sure we have a planet. This will also reset the planet to the
    startup values. */
void
GSimState::addPlanet()
{
    planet = GSimPlanet();
    if (!has_planet) {
        has_planet = true;
        structure_change = true;
    } else {
        planet.setChanged();
    }
}

/** Make sure we don't have a planet. */
void
GSimState::removePlanet()
{
    if (has_planet) {
        has_planet = false;
        structure_change = true;
    }
}

/** Perform all queued updates. */
void
GSimState::doUpdates()
{
    if (structure_change) {
        sig_structure_changed.raise();
        structure_change = false;
    }
    if (has_planet && planet.isChanged()) {
        sig_planet_changed.raise();
        planet.setUnchanged();
    }
    for (ship_index_t i = 0; i < ships.size(); ++i) {
        if (ships[i].isChanged()) {
            sig_ship_changed.raise(i);
            ships[i].setUnchanged();
        }
    }
}

/** Swap two states. */
void
GSimState::swap(GSimState& other)
{
    ships.swap(other.ships);
    std::swap(planet, other.planet);
    std::swap(has_planet, other.has_planet);
    structure_change = true;
    other.structure_change = true;
}

/** Assign random friendly codes for all objects that want them.
    Calls GSimObject::assignRandomFCode() on all objects. */
void
GSimState::assignRandomFCodes()
{
    for (ship_index_t i = 0; i < ships.size(); ++i)
        ships[i].assignRandomFCode();
    if (has_planet)
        planet.assignRandomFCode();
}

/** Check whether the specified ship Id is in use. */
bool
GSimState::isShipIdInUse(int sid) const
{
    for (ship_index_t i = 0; i < ships.size(); ++i)
        if (ships[i].getId() == sid)
            return true;
    return false;
}

/** Get ship by Id.
    \return pointer to ship object; 0 if not found. */
const GSimShip*
GSimState::getShipById(int sid) const
{
    for (ship_index_t i = 0; i < ships.size(); ++i)
        if (ships[i].getId() == sid)
            return &ships[i];
    return 0;
}

/** Get ship by Id.
    \return pointer to ship object; 0 if not found. */
GSimShip*
GSimState::getShipById(int sid)
{
    return const_cast<GSimShip*>(const_cast<const GSimState*>(this)->getShipById(sid));
}

/** Get planet. */
GSimPlanet&
GSimState::getPlanet()
{
    ASSERT(has_planet);
    return planet;
}

/** Get planet. */
const GSimPlanet&
GSimState::getPlanet() const
{
    ASSERT(has_planet);
    return planet;
}

/** Add a ship.
    \return index of new ship */
GSimState::ship_index_t
GSimState::addShip()
{
    ships.push_back(GSimShip());
    structure_change = true;
    return ships.size() - 1;
}

/** Get index of a GSimShip object.
    \param sh Some ship from this GSimState
    \inv getIndexOf(getShip(x)) == x */
GSimState::ship_index_t
GSimState::getIndexOf(const GSimShip& sh) const
{
    for (ship_index_t i = 0; i < ships.size(); ++i)
        if (&ships[i] == &sh)
            return i;
    return ships.size();
}

/** Get index of a ship by Id. */
GSimState::ship_index_t
GSimState::getIndexOf(int id) const
{
    for (ship_index_t i = 0; i < ships.size(); ++i)
        if (ships[i].getId() == id)
            return i;
    return ships.size();
}

/** Add ship at specific position.
    \param i ship to duplicate
    \param id Id number for the duplicate */
void
GSimState::duplicateShip(ship_index_t i, int id)
{
    ships.insert(ships.begin() + i + 1, 1, ships[i]);
    ships[i+1].setId(id);
    ships[i+1].setName(format(_("Ship %d"), id));
    structure_change = true;
}

/** Get number of ships. */
GSimState::ship_index_t
GSimState::getNumShips() const
{
    return ships.size();
}

/** Remove a ship.
    \param ix Index of ship to remove, [0,getNumShips()). */
void
GSimState::removeShip(ship_index_t ix)
{
    if (ix < ships.size()) {
        for (ship_index_t i = ix+1; i < ships.size(); ++i)
            ships[i-1] = ships[i];
        ships.pop_back();
        structure_change = true;
    }
}

/** Swap two ships.
    \param a,b Indexes of ships to swap */
void
GSimState::swapShips(ship_index_t a, ship_index_t b)
{
    GSimShip tmp = ships[a];
    ships[a] = ships[b];
    ships[b] = tmp;
    structure_change = true;
}

/** Get ship by index.
    \param sid Index of ship, [0,getNumShips()) */
GSimShip&
GSimState::getShip(ship_index_t sid)
{
    ASSERT(sid >= 0 && sid < ships.size());
    return ships[sid];
}

/** Get ship by index.
    \param sid Index of ship, [0,getNumShips()) */
const GSimShip&
GSimState::getShip(ship_index_t sid) const
{
    ASSERT(sid >= 0 && sid < ships.size());
    return ships[sid];
}

/** Clear all statistics counters embedded in the combat objects. */
void
GSimState::initializeStats()
{
    /* min_fighters_aboard needs some care: for ships, we have good reasons to initialize
       it to getAmmo() because that's the number of fighters initially on board.
       For planets, we'd need to duplicate the formulas here.
       Therefore, we just initialize it it INT_MAX, and filter it out in display
       (xref WSimUnitStat::render) */
    for (ship_index_t i = 0; i < ships.size(); ++i) {
        GSimShip&    sh = ships[i];
        VcrStatItem& it = sh.getStat();
        it.torps_hit = 0;
        it.min_fighters_aboard = sh.getAmmo();
        it.num_fights = 0;
    }
    if (has_planet) {
        planet.getStat().torps_hit = 0;
        planet.getStat().min_fighters_aboard = INT_MAX;
        planet.getStat().num_fights = 0;
    }
}

/** Get a free Id number.
    \return Id number, 0 if none found */
int
GSimState::getFreeId()
{
    for (int id = 1; id <= NUM_SHIPS; ++id) {
        if (!isShipIdInUse(id))
            return id;
    }
    return 0;
}

/** Check whether simulation matches ship list. Checks all ships.
    \see GSimShip::isMatchingShipList(). */
bool
GSimState::isMatchingShipList() const
{
    for (ship_index_t i = 0; i < ships.size(); ++i)
        if (!ships[i].isMatchingShipList())
            return false;
    return true;
}
