/**
  *  \file ui/table.cc
  *  \brief Display of "tables"
  *
  *  This module contains classes to display and measure "tabular"
  *  output. Mainly, this concerns things like planet spec listings,
  *  etc. Having this interface, changing fonts or NLS don't mess up
  *  our layouts.
  *
  *  To do some tabular output, write a function which takes an
  *  UIAbstractTableWriter and calls its methods. To use that output,
  *  you'll later call that method with an instance of UITableSizer or
  *  UITableWriter; maybe even another class. The table writer
  *  protocol was designed to require as little as possible temporary
  *  storage for the normal case. Actually, none of the current
  *  classes temporarily saves any cell value.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/skin.h"
#include "gfx/gfx.h"

/************************* UIAbstractTableWriter *************************/

/** \class UIAbstractTableWriter
    \brief Basic Table Writer Class

    This class provides the interface for displaying tables. Tables
    are displayed column-wise. To display a table, call newLColumn()
    or newRColumn(), and fill in cells with writeFixedCell() and
    writeVariableCell(). Repeat that cycle as long as you need.

    See these functions for details. See UITableWriter for some
    restrictions. */

/** Constructor. */
UIAbstractTableWriter::UIAbstractTableWriter()
    : write_flags(0), col_width(0), cur_line(0), cur_x(-COLUMN_SEPARATOR)
{ }

/** Finish the table. You should call that function after you've
    finished writing your table. */
void
UIAbstractTableWriter::finish()
{ }

/** Write fixed cell. Call to write cells that do not change. */
void
UIAbstractTableWriter::writeFixedCell(UISkin::Color c, const char* s)
{
    writeFixedCell(c, string_t(s));
}

/** Write variable cell. Call to write cells that do change. */
void
UIAbstractTableWriter::writeVariableCell(UISkin::Color c, const char* s)
{
    writeVariableCell(c, string_t(s));
}

/** Start a new left-justified column. All cells will be left-justified.
    \param min_size minimum width of the column. If needed, the column
    will grow. If omitted, the column will be as narrow as possible. */
void
UIAbstractTableWriter::newLColumn(int min_size)
{
    cur_x += COLUMN_SEPARATOR + col_width;
    col_width   = min_size;
    cur_line    = 0;
    write_flags = fl_LJust;
}

/** Start a new column extending up to a particular X. Same as
    newLColumn(), but uses (at least) space up to specified X. */
void
UIAbstractTableWriter::newLColumnUpTo(int end_x)
{
    newLColumn(end_x - cur_x - col_width - COLUMN_SEPARATOR);
}

/** Start a new right-justified column. All cells will be right-justified.
    \param width width of the column. All cells will have this width (in
    pixels), truncated if needed. */
void
UIAbstractTableWriter::newRColumn(int width)
{
    cur_x += COLUMN_SEPARATOR + col_width;
    col_width   = width;
    cur_line    = 0;
    write_flags = 0;
}

/** Generic cell writer (back-end to writeFixedCell and writeVariableCell).
    \param c color to use
    \param s text to write
    \param flags flags that describe the output */
void
UIAbstractTableWriter::writeCell(UISkin::Color c, const string_t& s, int flags)
{
    flags |= write_flags;
    if (flags & fl_LJust) {
        /* left-justified cell */
        int cell_width = writeAnyCell(c, s, flags, col_width);
        if (cell_width > col_width)
            col_width = cell_width;
        ++cur_line;
    } else {
        /* right-justified cell. These have a fixed size and don't
           change the cell witdh. */
        writeAnyCell(c, s, flags, col_width);
        ++cur_line;
    }
}

/** Handle cell content without output. Makes sure the current column
    can hold the given string without resizing. Only works for
    left-justified columns.
    \param c text */
void
UIAbstractTableWriter::sizeCell(string_t s)
{
    if (write_flags & fl_LJust) {
        int cell_width = sizeAnyCell(s);
        if (cell_width > col_width)
            col_width = cell_width;
    }
}

/****************************** UITableSizer *****************************/

/** \class UITableSizer
    \brief Table Size Measurement

    This class can be used to measure the expected size of a table.
    For example, to implement getLayoutInfo() for a tile, you could
    feed the table output function with a UITableSizer. */

/** Constructor.
    \param font_nr number of the font to use */
UITableSizer::UITableSizer(int font_nr)
    : fnt(*fonts[font_nr]), max_lines(0)
{ }

/** Constructor.
    \param fnt font to use */
UITableSizer::UITableSizer(GfxFont& fnt)
    : fnt(fnt), max_lines(0)
{ }

UITableSizer::~UITableSizer()
{ }

int
UITableSizer::writeAnyCell(UISkin::Color /*c*/, const string_t& s, int flags, int col_size)
{
    if (getCurrentLine() >= max_lines)
        max_lines = getCurrentLine()+1;

    if (flags & fl_LJust) {
        return fnt.getTextWidth(s);
    } else {
        return col_size;
    }
}

int
UITableSizer::sizeAnyCell(const string_t& s)
{
    return fnt.getTextWidth(s);
}

/** Get height of table, in pixels. */
int
UITableSizer::getHeight() const
{
    return max_lines * fnt.getTextHeight(string_t());
}

GfxPoint
UITableSizer::getCurrentPosition() const
{
    return GfxPoint(getCurrentX(), getCurrentLine() * fnt.getTextHeight(string_t()));
}

/***************************** UITableWriter *****************************/

/** \class UITableWriter
    \brief Graphical Table Output

    This table writer will display the table contents on a skinnable
    surface. It can also limit output to a particular area.

    Limitations: this class does not properly support left-justified
    columns which contain some fixed and some variable cells.
    Right-justified columns with mixed content, or columns with only
    one cell type work properly.

    \todo Output size limits are currently only enforced for
    left-justified cells; vertical output size limits are not enforced
    at all. */

/** Constructor.
    \param ctx     context
    \param limit   area of output. This defines the origin of the table,
                   as well as output size limits.
    \param font_nr font to use for output */
UITableWriter::UITableWriter(GfxContext& ctx, GfxRect limit, int font_nr)
    : ctx(ctx), font_nr(font_nr), ignore(ignore), limit(limit)
{
    ctx.useFont(*fonts[font_nr]);
}

UITableWriter::~UITableWriter()
{ }

void
UITableWriter::finish()
{
    const int right = getCurrentX() + getCurrentWidth();
    const int h = font_heights[font_nr]*getCurrentLine();
    if (right < limit.w)
        drawBackground(ctx, GfxRect(limit.x + right, limit.y,
                                    limit.w - right, limit.h));
    if (h < limit.h)
        drawBackground(ctx, GfxRect(limit.x, limit.y + h,
                                    limit.w, limit.h - h));
}

int
UITableWriter::sizeAnyCell(const string_t& s)
{
    return fonts[font_nr]->getTextWidth(s);
}

int
UITableWriter::writeAnyCell(UISkin::Color c, const string_t& s, int flags, int col_size)
{
    int text_width = fonts[font_nr]->getTextWidth(s);
    const int y = limit.y + font_heights[font_nr]*getCurrentLine();

    /* okay, draw it. */
    if (flags & fl_Solid)
        ctx.setSolidBackground();
    else
        ctx.setTransparentBackground();
    ctx.setColor(c);
    if (flags & fl_LJust) {
        /* left-justified cell */
        int remaining  = limit.w - getCurrentX();

        if (text_width < col_size)
            text_width = col_size;
        if (text_width > remaining)
            text_width = remaining;

        ctx.setTextAlign(0, 0);
        outTextF(ctx, limit.x + getCurrentX(), y, text_width, s);
        if (text_width > col_size) {
            // we have expanded a cell
            if (getCurrentLine() && (flags & fl_Solid))
                drawBackground(ctx, GfxRect(limit.x + getCurrentX() + col_size, limit.y,
                                            text_width - col_size, y - limit.y));
        }

        return text_width;
    } else {
        /* right-justified cell */
        ctx.setTextAlign(2, 0);
        outTextF(ctx, limit.x + getCurrentX() + col_size, y, col_size, s);
        return col_size;
    }
}

GfxPoint
UITableWriter::getCurrentPosition() const
{
    return GfxPoint(limit.x + getCurrentX(),
                    limit.y + getCurrentLine() * font_heights[font_nr]);
}
