/**
  *  \file util/conf.h
  *  \brief Configuration Data Management
  *
  *  Contains classes to manage configuration data intended to be
  *  stored in configuration files. A Config object will contain many
  *  ConfigOption descendants, each of which can be looked up by a
  *  string key. Code to actually load and save such objects to files
  *  is in util/confpars.h.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_CONF_H
#define PCC_V2_UTIL_CONF_H

#include <vector>
#include "cpluslib/string.h"
#include "cpluslib/assert.h"
#include "cpluslib/types.h"
#include "cpluslib/iterator.h"

class ConfigOption;

void removeComment(string_t& str);

/** Configuration object. An object of this type contains parser and
    value information for a configuration.

    \todo How do we make temporary copies of such an object and/or
    transfer data between two of them? */
class Config {
    typedef std::vector<ConfigOption*> container_t;
    container_t options;

    Config(const Config&);
    Config& operator=(const Config&);
 public:
    Config();
    ~Config();
    void addConfigOption(ConfigOption& option);
    void markAllOptionsUnset();
    ConfigOption* getConfigOption(string_t name);

    typedef dereferencing_iterator<container_t::const_iterator> iterator;
    iterator begin()
        { return const_cast<const container_t&>(options).begin(); }
    iterator end()
        { return const_cast<const container_t&>(options).end(); }
};

/** Configuration option, base class. This is the basic interface to
    assigning configuration options. */
class ConfigOption {
 public:
    enum Source {
        srcDefault,
        srcSystem,
        srcUser,
        srcGame
    };
    ConfigOption(Config& container, const char* name);
    virtual ~ConfigOption()
        { }
    virtual void set(string_t value) = 0;
    virtual string_t toString() const = 0;
    const char* getName() const;

    Source getSource() const;
    void setSource(Source src);
    bool wasSet() const;
    void markUpdated(Source src);
    void setAndMarkUpdated(string_t value, Source src);
 private:
    ConfigOption(const ConfigOption&);
    void operator=(const ConfigOption&);
    const char* name;
    Source source;
};

/** Value parser, base class. Descendants of this class parse strings
    into integers according to particular rules. */
class ValueParser {
 public:
    virtual ~ValueParser()
        { }
    virtual int32 parse(string_t value) = 0;
    virtual string_t toString(int32 value) = 0;
    void parseArray(string_t value, int32 array[], int dim);
    string_t toStringArray(const int32 array[], int dim);
};

/** Value parser for integers. This parses regular decimal integers. */
class ValueIntParser : public ValueParser {
 public:
    ValueIntParser();
    int32 parse(string_t value);
    string_t toString(int32 value);
    static ValueIntParser instance;
};

/** Value parser for booleans. This parses PHost-style bools (Yes/No/Allies/External). */
class ValueBoolParser : public ValueParser {
 public:
    ValueBoolParser();
    int32 parse(string_t value);
    string_t toString(int32 value);
    static ValueBoolParser instance;
};

/** Value parser for a set of bits. This parses sets of bits. */
class ValueBitsetParser : public ValueParser {
    const char* tpl;
 public:
    ValueBitsetParser(const char* tpl);
    int32 parse(string_t value);
    string_t toString(int32 value);
};

/** Value parser for an enum. */
class ValueEnumParser : public ValueParser {
    const char* tpl;
 public:
    ValueEnumParser(const char* tpl);
    int32_t parse(string_t value);
    string_t toString(int32_t value);
};

/** Integer option. This contains a single value, parsed from
    text according to a ValueParser. */
template<typename T>
class ConfigValueOption : public ConfigOption {
    T value;
    ValueParser& parser;
 public:
    ConfigValueOption(Config& container, const char* name, ValueParser& parser, T value);
    ConfigValueOption(Config& container, const char* name, ValueParser& parser);
    void set(string_t value);
    string_t toString() const;
    void set(int32 new_value)
        { value = new_value; }
    int32 operator()() const
        { return value; }
    template<typename R>
    void copyFrom(const ConfigValueOption<R>& other)
        { value = other.value; }
    ValueParser& getParser() const
        { return parser; }
};

/** Integer option. This contains a single int32_t value, parsed from
    text according to a ValueParser. */
typedef ConfigValueOption<int32_t> ConfigIntOption;

/** Integer reference option. Contains an int32_t stored elsewhere. */
typedef ConfigValueOption<int32_t&> ConfigIntRefOption;

/** Base integer option array. Base class for an int32_t array, can
    inform user about its dimensions. This is used for the script interface,
    but shouldn't be used directly. */
class ConfigIntArrayBaseOption : public ConfigOption {
 public:
    ConfigIntArrayBaseOption(Config& container, const char* name)
        : ConfigOption(container, name)
        { }
    virtual const int32_t* getArray() const = 0;
    virtual int getLimit() const = 0;
    bool isAllTheSame() const;
};

/** Integer option array. This contains an array of int32 values,
    parsed from a comma-separated list according to a ValueParser. */
template<int N>
class ConfigIntArrayOption : public ConfigIntArrayBaseOption {
    int32 values[N];
    ValueParser& parser;
 public:
    ConfigIntArrayOption(Config& container, const char* name, ValueParser& parser)
        : ConfigIntArrayBaseOption(container, name),
          parser(parser)
        { set(0); }
    ConfigIntArrayOption(Config& container, const char* name, ValueParser& parser, int32_t value)
        : ConfigIntArrayBaseOption(container, name),
          parser(parser)
        { set(value); }
    void set(string_t value);
    string_t toString() const;
    void set(int index, int value);
    void set(int value);
    int32 operator()(int index) const;
    void copyFrom(const ConfigIntArrayOption<N>& other);
    const int32_t* getArray() const;
    int getLimit() const;
    ValueParser& getParser() const
        { return parser; }
};

/** String option. This contains a single string. */
class ConfigStringOption : public ConfigOption {
    string_t value;
 public:
    ConfigStringOption(Config& container, const char* name);
    ~ConfigStringOption();
    void set(string_t value);
    string_t toString() const;
    string_t operator()() const;
};

/** Alias option. This simply forwards all requests to another option. */
class ConfigAlias : public ConfigOption {
    ConfigOption& other;
 public:
    ConfigAlias(Config& container, const char* name, ConfigOption& other);
    void set(string_t value);
    string_t toString() const;
    ConfigOption& getForwardedOption()
        { return other; }
};

/****************************** ConfigOption *****************************/

/** Get name of this option. */
inline const char*
ConfigOption::getName() const
{
    return name;
}

inline ConfigOption::Source
ConfigOption::getSource() const
{
    return source;
}

inline void
ConfigOption::setSource(Source src)
{
    this->source = src;
}

inline bool
ConfigOption::wasSet() const
{
    return source != srcDefault;
}

/*************************** ConfigValueOption ***************************/

/** Constructor.
    \param container Containe to put this option into
    \param name Name of option
    \param parser Parser that parses strings into integers
    \param value Initial value */
template<typename T>
ConfigValueOption<T>::ConfigValueOption(Config& container, const char* name, ValueParser& parser, T value)
    : ConfigOption(container, name),
      value(value),
      parser(parser)
{ }

template<typename T>
ConfigValueOption<T>::ConfigValueOption(Config& container, const char* name, ValueParser& parser)
    : ConfigOption(container, name),
      value(0),
      parser(parser)
{ }

template<typename T>
void
ConfigValueOption<T>::set(string_t value)
{
    removeComment(value);
    this->value = parser.parse(value);
}

template<typename T>
string_t
ConfigValueOption<T>::toString() const
{
    return parser.toString(value);
}

/************************** ConfigIntArrayOption *************************/

template<int N>
void
ConfigIntArrayOption<N>::set(string_t value)
{
    removeComment(value);
    parser.parseArray(value, this->values, N);
}

template<int N>
string_t
ConfigIntArrayOption<N>::toString() const
{
    return parser.toStringArray(this->values, N);
}

template<int N>
void
ConfigIntArrayOption<N>::set(int index, int value)
{
    ASSERT(index > 0 && index <= N);
    this->values[index-1] = value;
}

template<int N>
void
ConfigIntArrayOption<N>::set(int value)
{
    for (int i = 0; i < N; ++i)
        this->values[i] = value;
}

template<int N>
int32
ConfigIntArrayOption<N>::operator()(int index) const
{
    // if index is out of bounds, return last value (i.e. Colony). This makes
    // more sense then returning the first (Fed), which differs in more ways
    // from standard than colony, at least for options important to us.
    if(index > 0 && index <= N)
        return this->values[index-1];
    else
        return this->values[N-1];
}

template<int N>
void
ConfigIntArrayOption<N>::copyFrom(const ConfigIntArrayOption<N>& other)
{
    for (int i = 0; i < N; ++i)
        this->values[i] = other.values[i];
}

template<int N>
const int32_t*
ConfigIntArrayOption<N>::getArray() const
{
    return this->values;
}

template<int N>
int
ConfigIntArrayOption<N>::getLimit() const
{
    return N;
}

#endif
