/**
  *  \file util/plugin.h
  *  \brief Plugin
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_PLUGIN_H
#define PCC_V2_UTIL_PLUGIN_H

#include <map>
#include <vector>
#include "cpluslib/string.h"
#include "io/streams.h"

class Plugin {
 public:
    enum ItemType {
        PlainFile,              // Just a file to copy
        ScriptFile,             // A file to copy and Load()
        ResourceFile,           // A file to copy and LoadResource()
        Command                 // Not a file, just a command
    };
    struct Item {
        ItemType type;
        string_t name;
        Item(ItemType type, const string_t& name)
            : type(type), name(name)
            { }
    };
    typedef std::vector<Item> ItemList;
    typedef std::map<string_t, string_t> FeatureSet;
    
    Plugin(string_t id);
    ~Plugin();

    void initFromPluginFile(string_t baseDir, string_t defFileName, Stream& file);
    void initFromResourceFile(string_t baseDir, string_t resFileName);
    void initFromScriptFile(string_t baseDir, string_t scriptFileName, Stream& file);

    void savePluginFile(Stream& file) const;
    void setBaseDirectory(const string_t& baseDir);

    void addItem(ItemType type, const string_t& name);

    void setLoaded(bool flag);

    const string_t& getId() const;
    const string_t& getName() const;
    const string_t& getDescription() const;
    const string_t& getBaseDirectory() const;
    const string_t& getDefinitionFileName() const;
    const ItemList& getItems() const;
    bool isProvided(const string_t& feature) const;
    bool isConflict(const Plugin& other) const;
    bool isUpdateFor(const Plugin& other) const;
    bool isDependingOn(const Plugin& other) const;
    bool isSatisfied(const FeatureSet& fset) const;
    bool isLoaded() const;
    bool enumMissingFeatures(const FeatureSet& have, FeatureSet& missing) const;

    void addProvidedFeaturesTo(FeatureSet& fset);

 private:
    string_t id;
    string_t name;
    string_t description;
    string_t baseDir;
    string_t defFileName;
    FeatureSet provides;
    FeatureSet requires;
    ItemList items;
    bool loaded;
};

bool compareVersions(const string_t& a, const string_t& b);


inline void
Plugin::setLoaded(bool flag)
{
    this->loaded = flag;
}

inline const string_t&
Plugin::getId() const
{
    return id;
}

inline const string_t&
Plugin::getName() const
{
    return name;
}

inline const string_t&
Plugin::getDescription() const
{
    return description;
}

inline const string_t&
Plugin::getBaseDirectory() const
{
    return baseDir;
}

inline const string_t&
Plugin::getDefinitionFileName() const
{
    return defFileName;
}

inline const Plugin::ItemList&
Plugin::getItems() const
{
    return items;
}

inline bool
Plugin::isLoaded() const
{
    return loaded;
}

#endif
