/**
  *  \file util/strtab.h
  *  \brief String Tables
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_STRTAB_H
#define PCC_V2_UTIL_STRTAB_H

#include "cpluslib/string.h"

/** Base class for a string table. Describes a list of strings, and has operations
    to operate on them. */
class StringTable {
 public:
    /** Destructor. */
    virtual ~StringTable()
        { }

    /** Get string corresponding to an index. */
    virtual string_t get(int index) const = 0;

    /** Get string corresponding to an index. */
    string_t operator()(int index) const
        { return get(index); }

    /** Get first index that has a sensible mapping. Defaults to returning 0 */
    virtual int getFirstIndex() const;

    /** Get index after last one that has a sensible mapping. */
    virtual int getLastIndex() const = 0;

    /** Get next index that has a mapping different from n. getLastIndex() if none.
        Defaults to increasing n (conservative, but possibly inefficient way). */
    virtual int getNextIndex(int n) const;
};

/** Static string table. Table containing a list of strings described by a static
    <tt>const char*</tt> array. */
class StaticStringTable : public StringTable {
    const char*const* table;
    int count;
 public:
    StaticStringTable(const char*const* table, std::size_t count);

    StaticStringTable(const char*const* table);

    string_t get(int index) const;

    int getLastIndex() const;
};

/** Static string table. Table containing a list of strings described by a static
    <tt>const char*</tt> array. The strings are translated before being used. */
class StaticTranslatedStringTable : public StaticStringTable {
 public:
    StaticTranslatedStringTable(const char*const* table, std::size_t count);

    StaticTranslatedStringTable(const char*const* table);

    string_t get(int index) const;
};

/** Construct from array.
    \param table Array
    \param count Number of elements in array */
inline
StaticTranslatedStringTable::StaticTranslatedStringTable(const char*const* table, std::size_t count)
    : StaticStringTable(table, count)
{ }

/** Construct from NULL-terminated array.
    \param table Array */
inline
StaticTranslatedStringTable::StaticTranslatedStringTable(const char*const* table)
    : StaticStringTable(table)
{ }

#endif
